/**<pre>
 * Title: iSIM
 * Copyright: Copyright (c) 2002
 * Company:   University of Oregon Computer Science Dept.
 *
 * @author Jason Prideaux
 * @version 1.0
 * </pre>
 */

package transport;

import transport.protocol.*;
import agent.*;
import controller.*;

import java.io.*;
import java.net.*;
import java.util.*;



/* ===================================================================== */
/** This class is used for Unicasting a message to one user in wireless
 *  range.  It adds in timing delays based on ratio of max range/distance
 *  to simulate bandwidth loss and delay due to wireless signal
 *  degradation.
 *
 */
public class Unicast extends Thread{

	/** The message to send in this thread. */
	private UnicastMessage msg;

	/** The list of all access points in the simulation. */
	private Vector points = new Vector();

	/** The list of all users in the simulation. */
	private Vector users = new Vector();



	/* ================================================================ */
	/** This Costructor preps the Thread for the given message to send.
	 *
	 *  @param  msg   The message to send to other users in range.
	 *
	 */
	public Unicast(Vector users, Vector points, UnicastMessage msg){

		this.users = users;
		this.points = points;
		this.msg = msg;

	} //Constructor



	/* ================================================================ */
	/** This method excutes the thread.  It sends the message to one
	 *  Users in range, adds a delay, and sends to the Users if it can be
	 *  reached in a number of hops specified by the message.
	 *
	 */
	public void run(){

		System.out.println("Processing a Unicast message...");

		int hops = msg.getHops();
		int from = msg.getFrom();
		int to = msg.getTo();

		double diff = 0;
		Agent a = null;
		Agent b = null;
		Agent temp = null;
		Vector routers = new Vector();

		// Find the agent that sent the message.
		for( int i=0; i <users.size(); i++ ){
			temp = (Agent)users.elementAt(i);
			if( from == ((User)temp).getPort() ){
				a = temp;
			}
		}

		// Find the agent that the message will go to.
		for( int i=0; i <users.size(); i++ ){
			temp = (Agent)users.elementAt(i);
			if( to == ((User)temp).getPort() ){
				b = temp;
			}
		}

		routers.add(a);

		System.out.println("   sent by "+ ((User)a).getPort() + "   going to "+ ((User)b).getPort());

		try{

		//Now send message to only one agent if in range.
		if( a != null && b != null){

			// Send to all agents in one hop.
			for( int i=0; i <users.size(); i++ ){

				temp = (Agent)users.elementAt(i);
				diff = Controller.diff(a.getLocation(),temp.getLocation());

				//System.out.println( "diff: "+diff);

				if( (diff < (a.getRange()+temp.getRange())) && !(routers.contains(temp)) ){;

					routers.add(temp);
				}
			}

			for( int i=0; i <points.size(); i++ ){
				temp = (Agent)points.elementAt(i);
				diff = Controller.diff(a.getLocation(),temp.getLocation());

				if( diff < (a.getRange()+temp.getRange()) && !(routers.contains(temp)) ){
					routers.add(temp);
				}
			}


			// Now find those within hops.
			int h = 1;
			while( h < hops ){
				h++;

				for( int j=0; j < routers.size(); j++ ){

					a = (Agent)routers.elementAt(j);

					// Find all agents in one hop.
					for( int i=0; i <users.size(); i++ ){

						temp = (Agent)users.elementAt(i);
						diff = Controller.diff(a.getLocation(),temp.getLocation());

						if( !(routers.contains(temp)) && diff < (a.getRange()+temp.getRange()) ){

							routers.add(temp);
						}
					}
					for( int i=0; i <points.size(); i++ ){

						temp = (Agent)points.elementAt(i);
						diff = Controller.diff(a.getLocation(),temp.getLocation());

						if( !(routers.contains(temp)) && diff < (a.getRange()+temp.getRange()) ){
							routers.add(temp);
						}
					}

				}
			} //while


			//If the agent can be reached then send the message
			if( routers.contains(b) ){

				// Impose a slight delay
				try{
					Thread.sleep(750);
				}catch(Exception e){}

				System.out.println( "      sending message to: "+to);

				if( a.isWirelessSensor() && b.isWirelessSensor() && a.isWirelessSensor2() && b.isWirelessSensor2()){
					SocketSender ss = new SocketSender(msg);
					ss.start();
				}else if( !b.isWirelessSensor() ){
					SocketSender ss = new SocketSender(new ErrorMessage(((User)b).getPort(),2323, msg));
					ss.start();
				}else if( !a.isWirelessSensor() ){
					SocketSender ss = new SocketSender(new ErrorMessage(((User)a).getPort(),2323, msg));
					ss.start();
				}
			}

		}

		}catch(Exception e){System.out.println("Error in message sending method of unicast "+e);}

	} //method: run

} //class: Unicast
